import tensorflow as tf
# import wandb

# TODO now deprecated
class UpdateStepCallback(tf.keras.callbacks.Callback):

    def __init__(self, batch_size):
        """
        I couldn't work out the "proper" way to track steps within a Sequential model for tensorboard, so this is a workaround.py
        Use a custom callback using keras.backend.set_value to set self.epoch=epoch, and then read that on each summary call
        Model must already have self.epoch = tf.Variable(0).
        https://www.tensorflow.org/guide/keras/custom_callback

        Args:
            batch_size (int): batch size, used to calculate steps as batch_size * epoch
        """
        super(UpdateStepCallback, self).__init__()
        self.batch_size = batch_size

    def on_epoch_end(self, epoch, logs=None):
        # print('\n\nStarting epoch', epoch, '\n\n')
        # # access model with self.model, tf.ketas.backend.get/set_value
        # # e.g. lr = float(tf.keras.backend.get_value(self.model.optimizer.lr))

        # print('\n epoch ', epoch, type(epoch))
        step = epoch * self.batch_size
        # # self.model.step = step
        # # self.model.step.assign(step)
        tf.keras.backend.set_value(self.model.step, step)
        # print('\n Ending step: ', float(tf.keras.backend.get_value(self.model.step)))
        # # print(f'Step {step}')


# class VisualizeImages(tf.keras.callbacks.Callback):

#     # no args needed yet
#     # def __init__(self):
#     #     super(VisualizeImages, self).__init__()

#     def on_train_begin(self, logs=None):
#         tf.summary.image(name='images_before_augmentation', data, max_outputs=3, description='Images passed to Zoobot')


# callbacks can use self.model
# https://www.tensorflow.org/guide/keras/custom_callback#usage_of_selfmodel_attribute


import logging

from lightning_lite.plugins.environments import SLURMEnvironment

# https://pytorch-lightning.readthedocs.io/en/stable/_modules/lightning_lite/plugins/environments/slurm.html#SLURMEnvironment
# https://github.com/Lightning-AI/lightning/blob/9c20cad40e4142f8a5e945fe26e919e598f2bd56/src/lightning_lite/plugins/environments/slurm.py
class ManchesterEnvironment(SLURMEnvironment):

    def __init__(self, auto_requeue: bool = True, requeue_signal=None) -> None:
        logging.info('Using Manchester SLURM environment')
        super().__init__(auto_requeue, requeue_signal)

    # @staticmethod
    # def resolve_root_node_address(nodes: str) -> str:
    #     root_node_address = super().resolve_root_node_address(nodes)
    #     logging.info(f'root_node_address: {root_node_address}')
    #     return root_node_address

    @staticmethod
    def detect() -> bool:
        return True
        
    @property
    def main_port(self) -> int:
        main_port = super().main_port
        logging.info(f'main_port: {main_port}')
        return main_port
        # MASTER_PORT will override

if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)

    # slurm_nodelist = "compute-0-[0,9]" # 0,9 works
    slurm_nodelist = "compute-0-[0,11]"  # 0,11 hangs
    # 70017 8-9 works

    env = ManchesterEnvironment()
    root = env.resolve_root_node_address(slurm_nodelist)
    print(root)

    print(env.detect())

    print(env.main_port)


import os

import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
import tqdm
import pandas as pd
from sklearn.decomposition import IncrementalPCA


def create_pca_embedding(features: np.array, n_components: int, variance_plot_loc=None):
    """
    Compress galaxy representations into a lower dimensionality using Incremental PCA.
    These compressed representations are easier and faster to work with.

    Args:
        features (np.array): galaxy representations, of shape (galaxies, feature_dimensions)
        n_components (int): number of PCA components to use. Sets output dimension.
        variance_plot_loc (str, optional): If not None, save plot of variance vs. PCA components here. Defaults to None.

    Raises:
        ValueError: features includes np.nan values (PCA would break)

    Returns:
        np.array: PCA-compressed representations, of shape (galaxies, pca components)
    """
    assert len(features) > 0
    pca = IncrementalPCA(n_components=n_components, batch_size=20000)
    reduced_embed = pca.fit_transform(features)
    if np.isnan(reduced_embed).any():
        raise ValueError(f'embed is {np.isnan(reduced_embed).mean()} nan')

    if variance_plot_loc:
        plt.plot(range(n_components), pca.explained_variance_ratio_)
        plt.xlabel('Nth Component')
        plt.ylabel('Explained Variance')
        plt.tight_layout()
        plt.savefig(variance_plot_loc)

    return pd.DataFrame(data=reduced_embed, columns=['feat_{}_pca'.format(n) for n in range(n_components)])

"""
The code below is deprecated because I no longer use dataframes for this, but rather .hdf5. However, the approach might be useful to read."""

# def main(df: pd.DataFrame, name: str, output_dir: str, components_to_calculate=[5, 10, 30], id_col='iauname'):
#     """
#     Wrapper around :meth:`create_pca_embedding`.
#     Creates and saves several embeddings using (by default) 5, 10, and 30 PCA components.

#     Args:
#         df (pd.DataFrame): with columns of id_col (below) and feat_* (e.g. feat_0_pred, feat_1_pred, ...) recording representations for each galaxy (row)
#         name (str): Text to identify saved outputs. No effect on results.
#         output_dir (str): Directory in which to save results. No effect on results.
#         id_col (str, optional): Name of column containing unique strings identifying each galaxy. Defaults to 'iauname', matching DECaLS catalog. 'id_str' may be useful to match GZ2 catalog.
#     """
#     feature_cols = [col for col in df.columns.values if col.startswith('feat')]
    
#     features = df[feature_cols].values
    
#     for n_components in tqdm.tqdm(components_to_calculate):

#         # only need to bother with the variance plot for the highest num. components
#         if n_components == np.max(components_to_calculate):
#             variance_plot_loc = os.path.join(output_dir, name + '_explained_variance.pdf')
#         else:
#             variance_plot_loc = None

#         embed_df = create_pca_embedding(features, n_components, variance_plot_loc=variance_plot_loc)
#         embed_df[id_col] = df[id_col]  # pca embedding doesn't shuffle, so can copy the id col across to new df
#         embed_df.to_parquet(os.path.join(output_dir, '{}_pca{}_and_ids.parquet'.format(name, n_components)), index=False)


# if __name__ == '__main__':

#     sns.set_context('notebook')

#     output_dir = '/Users/walml/repos/zoobot/data/results'
#     assert os.path.isdir(output_dir)

#     name = 'decals_dr5_oct_21' 
#     features_loc = '/Volumes/beta/cnn_features/decals/cnn_features_decals.parquet'  # TODO point this to your representations download from Zenodo
#     df = pd.read_parquet(features_loc)
#     # TODO replace second arg with your image download folder
#     df['png_loc'] = df['png_loc'].str.replace('/media/walml/beta1/decals/png_native/dr5', '/Volumes/beta/decals/png_native/dr5')  
#     id_col = 'iauname'

#     """
#     Similarly for GZ2
#     """
#     # name = 'gz2' 
#     # features_loc = '/Volumes/beta/cnn_features/gz2/cnn_features_gz2.parquet'  # TODO point this to your download from Zenodo
#     # df = pd.read_parquet(features_loc)
#     # # TODO replace second arg with your image download folder
#     # df['png_loc'] = df['png_loc'].str.replace('/media/walml/beta1/galaxy_zoo/gz2/png', '/Volumes/beta/galaxy_zoo/gz2/png')  
#     # id_col = 'id_str'

#     main(df, name, output_dir, id_col=id_col)



import logging

import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow_probability as tfp


def get_hpd(x: np.ndarray, p: np.ndarray, ci=0.8):
    """
    Get highest posterior density interval containing roughly "ci" specified total prob.


    Args:
        x (np.ndarray): Values of random variable e.g. votes [0, 1, ..., N]
        p (np.ndarray): Discrete prob. of those values. Normalised to 1.
        ci (float, optional): [description]. Defaults to 0.8.

    Returns:
        (Any) Value of x at low edge of interval
        (Any) Value of x at high edge of interval
        (float) Total probability between edges (inclusive)
        (bool) True if posterior is unimodal 
    """

    if len(p) <= 1:
        print(x, p)
        raise IndexError
    assert x.ndim == 1
    assert x.shape == p.shape
    assert np.isclose(p.sum(), 1, atol=0.001)
    # here, x is discrete posterior p's, not samples as with agnfinder
    mode_index = np.argmax(p)
    # check unimodal
    unimodal = True
    if not np.argsort(p)[::-1][1] in (mode_index-1, mode_index+1):
        logging.warning(f'Possible second mode, hpd will fail: {p}')
        unimodal = False
    lower_index = mode_index
    higher_index = mode_index
    while True:
        confidence = p[lower_index:higher_index+1].sum()
        if confidence >= ci:
            break  # discrete so will be at least a little over
        else:  # step each boundary outwards towards the edge, stop at the edge
            lower_index = max(0, lower_index-1)
            higher_index = min(len(x)-1, higher_index+1)

    # these indices will give a symmetric interval of at least ci, but not exactly - and will generally overestimate
    # hence confidence will generally be a bit different to desired ci, important to return
    return (x[lower_index], x[higher_index]), confidence, unimodal
        

def get_coverage(posteriors, true_values):
    results = []
    for ci_width in np.linspace(0.1, 0.95):  # 50, by default
        for target_n, (x, posterior) in enumerate(posteriors):  # n posteriors
            true_value = true_values[target_n]
            (lower_lim, higher_lim), confidence, unimodal = get_hpd(x, posterior, ci=ci_width)
            within_any_ci = lower_lim <= true_value <= higher_lim  # inclusive
            results.append({
                'target_index': target_n,
                'requested_ci_width_dont_use': ci_width, # requested confidence
                'confidence': confidence,  # actual confidence, use this
                'lower_edge': lower_lim,
                'upper_edge': higher_lim,
                'true_value': true_value,
                'true_within_hpd': within_any_ci,
                'unimodal': unimodal
            })
    df = pd.DataFrame(results)
    df = df.drop_duplicates(subset=['target_index', 'confidence'])
    return df


def get_true_values(catalog, id_strs, answer):
    true_values = []
    for id_str in id_strs:
        galaxy = catalog[catalog['id_str'] == id_str].squeeze()
        true_values.append(galaxy[answer.text])
    return true_values


def get_posteriors(samples, catalog, id_strs, question, answer, temperature=None):
    """
    Samples and catalog are aligned - so let's just use them as one dataframe instead of two args


    Args:
        samples (np.ndarray): Corresponding prediction for galaxy with ``id_str``
        catalog (pd.DataFrame): Used for total votes.
        id_strs (list): [description]
        question (schemas.Question): [description]
        answer (schemas.Answer): [description]
        temperature (int, optional): Optional annealing of posteriors. Not recommended. None by default.

    Returns:
        list: posteriors like [yes_votes_arr, p_of_each] for each galaxy
    """
    all_galaxy_posteriors = []
    for sample_n, sample in enumerate(samples):
        galaxy = catalog[catalog['id_str'] == id_strs[sample_n]].squeeze()
        galaxy_posteriors = get_galaxy_posteriors(sample, galaxy, question, answer)
        all_galaxy_posteriors.append(galaxy_posteriors)
    if temperature is not None:
        all_galaxy_posteriors = [(indices, (posterior ** temperature) / np.sum(posterior ** temperature, axis=1, keepdims=True)) for (indices, posterior) in all_galaxy_posteriors]
    return all_galaxy_posteriors


def get_galaxy_posteriors(sample, galaxy, question, answer):
    assert answer in question.answers
    n_samples = sample.shape[-1]
    cols = [a.text for a in question.answers]
    assert len(cols) == 2 # Binary only!
    total_votes = galaxy[cols].sum().astype(np.float32)

    votes = np.arange(0., total_votes+1)
    x = np.stack([votes, total_votes-votes], axis=-1)  # also need the counts for other answer, no. 
    votes_this_answer = x[:, answer.index - question.start_index]  # second index is 0 or 1
    
    # could refactor with new equal mixture class
    all_probs = []
    for d in range(n_samples):
        concentrations = tf.constant(sample[question.start_index:question.end_index+1, d].astype(np.float32))
        probs = tfp.distributions.DirichletMultinomial(total_votes, concentrations).prob(x)
        all_probs.append(probs)
        
    return votes_this_answer, np.array(all_probs)




import json
from typing import List
import logging

import numpy as np
import pandas as pd
import h5py


def predictions_to_hdf5(predictions, id_str, label_cols, save_loc, compression="gzip"):
    logging.info(f'Saving predictions to {save_loc}')
    assert save_loc.endswith('.hdf5')
    with h5py.File(save_loc, "w") as f:
        f.create_dataset(name='predictions', data=predictions, compression=compression)
        # https://docs.h5py.org/en/stable/special.html#h5py.string_dtype
        dt = h5py.string_dtype(encoding='utf-8')
        # predictions_dset.attrs['label_cols'] = label_cols  # would be more conventional but is a little awkward
        f.create_dataset(name='id_str', data=id_str, dtype=dt)
        f.create_dataset(name='label_cols', data=label_cols, dtype=dt)
        # sometimes throws a "could not lock file" error but still saves fine. I don't understand why


def predictions_to_csv(predictions, id_str, label_cols, save_loc):
    # not recommended - hdf5 is much more flexible and pretty easy to use once you check the package quickstart
    assert save_loc.endswith('.csv')
    data = [prediction_to_row(predictions[n], id_str[n], label_cols=label_cols) for n in range(len(predictions))]
    predictions_df = pd.DataFrame(data)
    # logging.info(predictions_df)
    predictions_df.to_csv(save_loc, index=False)


def prediction_to_row(prediction: np.ndarray, id_str: str, label_cols: List):
    """
    Convert prediction on image into dict suitable for saving as csv
    Predictions are encoded as a json e.g. "[1., 0.9]" for 2 repeat predictions on one galaxy
    This makes them easy to read back with df[col] = df[col].apply(json.loads)

    Args:
        prediction (np.ndarray): model output for one galaxy, including repeat predictions e.g. [[1., 0.9], [0.3, 0.24]] for model with output_dim=2 and 2 repeat predictions
        id_str (str): path to image
        label_cols (list): semantic labels for model output dim e.g. ['smooth', 'bar'].

    Returns:
        dict: of the form {'id_str': 'path', 'smooth_pred': "[1., 0.9]", 'bar_pred: "[0.3, 0.24]"}
    
    """
    row = {
        'id_str': id_str  # may very well be a path to an image, if using an image dataset - just rename later
    }
    for n in range(len(label_cols)):
        answer = label_cols[n]
        answer_pred = prediction[n].astype(float)  # (n_samples,) shape
        if isinstance(answer_pred, float) or isinstance(answer_pred, np.float64):
            row[answer + '_pred'] = answer_pred  # it's a scalar already, life is good
        elif len(answer_pred) == 1:  # i.e. if only one sample
            row[answer + '_pred'] = answer_pred.squeeze()  # it's a scalar in disguise, make it a scalar 
        else:
            row[answer + '_pred'] = json.dumps(list(answer_pred))  # it's not a scalar, write as json
    return row



import os
os.environ['TF_XLA_FLAGS'] = '--tf_xla_auto_jit=-1'
import logging
import contextlib
from sklearn.model_selection import train_test_split

import tensorflow as tf
import wandb  # for direct manual hparam logging
from zoobot.tensorflow.training import training_config, losses, custom_metrics
from zoobot.tensorflow.estimators import define_model
from galaxy_datasets.tensorflow import get_image_dataset, add_transforms_to_dataset
from galaxy_datasets.transforms import default_transforms


def train(
    # absolutely crucial arguments
    save_dir,  # save model here
    schema,  # answer these questions
    # input data - specify *either* catalog (to be split) or the splits themselves
    catalog=None,
    train_catalog=None,
    val_catalog=None,
    test_catalog=None,
    # model training parameters
    architecture_name='efficientnet',  # only EfficientNet is currenty implemented
    batch_size=128,
    dropout_rate=0.2,
    # TODO drop_connect_rate not implemented
    epochs=1000,
    patience=8,
    # augmentation parameters
    color=False,
    requested_img_size=None,
    crop_scale_bounds=(0.7, 0.8),
    crop_ratio_bounds=(0.9, 1.1),
    resize_after_crop=224,
    always_augment=False,  # TODO deprecated for now
    # TODO specify test_time_dropout
    # hardware parameters
    mixed_precision=True,
    gpus=2,
    eager=False,  # tf-specific. Enable eager mode. Set True for easier debugging but slower training
    check_valid_paths=True,  # checks all images exist. Can disable for start speed on large datasets (100k+)
    # replication parameters
    random_state=42  # TODO not yet implemented except for catalog split (not used in benchmarks)
) -> tf.keras.Model:
    """
    Train a Zoobot (EfficientNetB0) model from scratch.
    Note that the TensorFlow Zoobot version is not actively developed.
    We suggest the PyTorch version instead.

    Args:
        save_dir (str): directory to save the trained model and logs
        train_catalog (pd.DataFrame, optional): Catalog of galaxies for training. Must include `id_str`, `file_loc`, and label_cols columns. Defaults to None.
        val_catalog (pd.DataFrame, optional): As above, for validation. Defaults to None.
        test_catalog (pd.DataFrame, optional): As above, for testing. Defaults to None.
        architecture_name (str, optional): Specifies architecture to use. B0 by default. Defaults to 'efficientnet'.
        dropout_rate (float, optional): Prob. of dropout prior to output layer. Defaults to 0.2.
        epochs (int, optional): Max epochs to train. Defaults to 1000.
        patience (int, optional): Max epochs to wait for loss improvement before cancelling training. Defaults to 8.
        color (bool, optional): Train with RGB images. Defaults to False.
        requested_img_size (int, optional): Size to load images from disk (i.e. to resize before transforms). Defaults to None.
        crop_scale_bounds (tuple, optional): Zoom fraction for random crops. Defaults to (0.7, 0.8).
        crop_ratio_bounds (tuple, optional): Aspect ratio for random crops. Defaults to (0.9, 1.1).
        resize_after_crop (int, optional): Size to input images to network (i.e. to resize after transforms). Defaults to 224.
        always_augment (bool, optional): Use augmentations at test time. Defaults to False.
        gpus (int, optional): Num. gpus to use. Defaults to 2.
        eager (bool, optional): Use eager execution (good for debugging but much slower). Defaults to False.

    Returns:
        tf.keras.Model: Trained Zoobot model.
    """

    # get the image paths, divide into train/val/test if not explicitly passed above
    if catalog is not None:
        assert train_catalog is None
        assert val_catalog is None
        assert test_catalog is None
        train_catalog, hidden_catalog = train_test_split(catalog, train_size=0.7, random_state=random_state)
        val_catalog, test_catalog = train_test_split(hidden_catalog, train_size=1./3., random_state=random_state)
    else:
        assert catalog is None
        assert len(train_catalog) > 0
        assert len(val_catalog) > 0
        assert len(test_catalog) > 0

    # a bit awkward, but I think it is better to have to specify you def. want color than that you def want greyscale
    greyscale = not color
    if greyscale:
        logging.info('Converting images to greyscale before training')
        channels = 1
    else:
        logging.warning(
            'Training on color images, not converting to greyscale')
        channels = 3

    assert save_dir is not None
    if not os.path.isdir(save_dir):
        os.mkdir(save_dir)

    if gpus > 1:
        logging.info('Using distributed mirrored strategy')
        strategy = tf.distribute.MirroredStrategy()  # one machine, one or more GPUs
        # strategy = tf.distribute.MultiWorkerMirroredStrategy()  # one or more machines. Not tested - you'll need to set this up for your own cluster.
        context_manager = strategy.scope()
        logging.info('Replicas: {}'.format(strategy.num_replicas_in_sync))
        # each GPU will calculate loss (hence gradients) for that device's sub-batch
        # within that sub-batch, loss uses tf.keras.reduction setting
        # I chose NONE (i.e. no reduction) and do subbatch_loss/full_batch_size (not sub-batch size)
        # this means the absolute loss and summed gradients don't change with more devices
        # TODO blog post on this    
        assert strategy.num_replicas_in_sync == gpus
    else:
        logging.info('Using single or no GPU, not distributed')
        # does nothing, just a convenience for clean code
        context_manager = contextlib.nullcontext()

    train_image_paths = list(train_catalog['file_loc'])
    val_image_paths = list(val_catalog['file_loc'])
    test_image_paths = list(test_catalog['file_loc'])

    # format is [{label_col: 0, label_col: 12}, {label_col: 3, label_col: 14}, ...]
    train_labels = train_catalog[schema.label_cols].values
    val_labels = val_catalog[schema.label_cols].values
    test_labels = test_catalog[schema.label_cols].values

    logging.info('Example path: {}'.format(train_image_paths[0]))
    logging.info('Example labels: {}'.format(train_labels[0]))

    logging.info(f'Will check if paths valid: {check_valid_paths}')
    train_dataset = get_image_dataset(
        train_image_paths, labels=train_labels, requested_img_size=requested_img_size, check_valid_paths=check_valid_paths, greyscale=greyscale
    )
    val_dataset = get_image_dataset(
        val_image_paths, labels=val_labels, requested_img_size=requested_img_size, check_valid_paths=check_valid_paths, greyscale=greyscale
    )
    test_dataset = get_image_dataset(
        test_image_paths, labels=test_labels, requested_img_size=requested_img_size, check_valid_paths=check_valid_paths, greyscale=greyscale
    )

    # specify augmentations
    train_transforms = default_transforms(
        # no need to specify greyscale here
        # tensorflow will greyscale in get_image_dataset i.e. on load, while pytorch doesn't so needs specifying here
        # may refactor to avoid inconsistency 
        crop_scale_bounds=crop_scale_bounds,
        crop_ratio_bounds=crop_ratio_bounds,
        resize_after_crop=resize_after_crop
    )
    # TODO should be clearer, not magic 1's (e.g. default_train_transforms, etc.)
    # TODO if always_augment, use train augments at test time (TODO rename this too)
    inference_transforms = default_transforms(
        crop_scale_bounds=(1., 1.),
        crop_ratio_bounds=(1., 1.),
        resize_after_crop=resize_after_crop
    )
    # apply augmentations
    train_dataset = add_transforms_to_dataset(train_dataset, train_transforms)
    # if always_augment:
        # logging.warning('always_augment=True, applying augmentations to val and test datasets')
    val_dataset = add_transforms_to_dataset(val_dataset, inference_transforms)
    test_dataset = add_transforms_to_dataset(test_dataset, inference_transforms)

    # batch, shuffle, prefetch
    train_dataset = train_dataset.shuffle(5000).batch(batch_size).prefetch(tf.data.experimental.AUTOTUNE)
    val_dataset = val_dataset.batch(batch_size).prefetch(tf.data.experimental.AUTOTUNE)
    test_dataset = test_dataset.batch(batch_size)

    with context_manager:

        if mixed_precision:
            logging.info(
                'Using mixed precision. \
                Note that this is a global setting (why, tensorflow, why...) \
                and so may affect any future code'
            )
            tf.keras.mixed_precision.set_global_policy('mixed_float16')

        # TODO use architecture_name here

        model = define_model.get_model(
            output_dim=len(schema.label_cols),
            input_size=resize_after_crop,
            channels=channels,
            dropout_rate=dropout_rate
        )

        # reduction=None will give per-example loss. Still summed (probability-multiplied) across questions.
        multiquestion_loss = losses.get_multiquestion_loss(
            schema.question_index_groups,
            sum_over_questions=True,
            reduction=tf.keras.losses.Reduction.NONE
        )
        # NONE reduction over loss, to be clear about how it's reduced vs. batch size
        # get the average loss, averaging over subbatch size (as TF *sums* gradients)
        def loss(x, y): return tf.reduce_sum(multiquestion_loss(x, y)) / (batch_size/gpus)        
        """
        TF actually has a built-in for this which just automatically gets num_replicas and does 

        per_replica_batch_size = per_example_loss.shape[0]
        global_batch_size = per_replica_batch_size * num_replicas
        return reduce_sum(per_example_loss) / global_batch_size

        but it's simple enough here that I'll just do it explicitly
        """
        # def loss(x, y): return tf.nn.compute_average_loss(per_example_loss=multiquestion_loss(x, y), global_batch_size=batch_size)  


        # be careful to define this within the context_manager, so it is also mirrored if on multi-gpu
        extra_metrics = [
            # this currently only works on 1 GPU - see Keras issue
            # custom_metrics.LossPerQuestion(
            #     name='loss_per_question',
            #     question_index_groups=schema.question_index_groups
            # )
        ]

    # https://docs.wandb.ai/guides/track/config#efficient-initialization
    if wandb.run is not None:  # user might not be using wandb
        wandb.config.update({
            'random_state': random_state,
            'epochs': epochs,
            'gpus': gpus,
            'precision': mixed_precision,
            'batch_size': batch_size,
            'greyscale': not color,
            'crop_scale_bounds': crop_scale_bounds,
            'crop_ratio_bounds': crop_ratio_bounds,
            'resize_after_crop': resize_after_crop,
            'framework': 'tensorflow',
            # tf doesn't automatically log model init args
            'architecture_name': architecture_name,  # only EfficientNet is currenty implemented
            'batch_size': batch_size,
            'dropout_rate': dropout_rate,
            # TODO drop_connect_rate not implemented
            'epochs': epochs,
            'patience': patience
        })


    model.compile(
        loss=loss,
        optimizer=tf.keras.optimizers.Adam(learning_rate=1e-3, beta_1=0.9, beta_2=0.999),
        metrics=extra_metrics,
        jit_compile=False  # don't use XLA, it fails on multi-GPU. Might consider on one GPU.
    )
    model.summary()

    trainer = training_config.Trainer(
        # parameters for how to train e.g. epochs, patience
        log_dir=save_dir,
        epochs=epochs,
        patience=patience
    )

    best_trained_model = trainer.fit(
        model,
        train_dataset,
        val_dataset,
        test_dataset,
        eager=eager,
        verbose=1
    )

    return best_trained_model



import logging
import os
from typing import List

import numpy as np
import pandas as pd
import h5py

from zoobot.shared import stats, schemas


def load_hdf5s(hdf5_locs: List):
    """
    Load hdf5 predictions (with metadata) from disk.
    
    Each hdf5 includes the following datasets (accessed with f[dataset_name]):
        - predictions: np.array of shape (galaxy, answer, forward pass) containing model outputs (usually Dirichlet concentrations)
        - id_str: unique string (usually the file location) specifying which galaxy each prediction refers to. Indices match `predictions`
        - label_cols: semantic names for the answer dimension of `predictions` e.g. smooth-or-featured-dr8_smooth, smooth-or-featured-dr8_featured-or-disk, ...
    
    See predictions.predictions_to_hdf5 for the save function.

    hdf5 is useful because predictions are three-dimensional (e.g. 10 galaxies, 4 answers, 5 dropout passes)
    while csv's and similar are designed for two-dimensional data (e.g 10 galaxies, 4 answers).

    Args:
        hdf5_locs (list): hdf5 files to load

    Returns:
        pd.DataFrame: with rows of id_str, hdf5_loc, indexed like predictions (below)
        np.array: model predictions, usually dirichlet concentrations, like (galaxy, answer, forward pass)
        list: semantic names of answers (e.g. ['smooth-or-featured-dr8_smooth', ...])
    """

    if isinstance(hdf5_locs, str):
        logging.warning('Passed a single hdf5 loc to load_hdf5s - assuming this is a single file to load, not list of files to load')
        hdf5_locs = [hdf5_locs]  # pretend user passed a list

    predictions = []
    prediction_metadata = []
    template_label_cols = None  # will use this var to check consistency of label_cols across each hdf5_loc
    for loc in hdf5_locs:
        with h5py.File(loc, 'r') as f:
            logging.debug(f.keys())
            these_predictions = f['predictions'][:]
            these_prediction_metadata = {
                'id_str': f['id_str'].asstr()[:],
                'hdf5_loc': [os.path.basename(loc) for _ in these_predictions]
        }
            predictions.append(these_predictions)  # will create a list where each element is 3D predictions stored in each hdf5
            prediction_metadata.append(these_prediction_metadata)  # also track id_str, similarly

            if template_label_cols is None:  # first file to load, use this as the expected template
                template_label_cols = f['label_cols'].asstr()[:]
                logging.info('Using label columns {} from first hdf5 {}'.format(template_label_cols, loc))
            else:
                these_label_cols = f['label_cols'].asstr()[:]
                if any(these_label_cols != template_label_cols):
                    raise ValueError('Label columns {} of hdf5 {} do not match first label columns {}'.format(loc, f['label_cols'], template_label_cols))


    # there is no assumption that id_str is unique, or attempt to group predictions by id_str
    # it just maps a set of hdf5 files, each with predictions, to a df of id_str and those loaded predictions (matching row-wise)
    logging.info('All hdf5 loaded, beginning concat.')
    predictions = np.concatenate(predictions, axis=0)
    prediction_metadata = {
        'id_str': [p for metadata in prediction_metadata for p in metadata['id_str']],
        'hdf5_loc': [l for metadata in prediction_metadata for l in metadata['hdf5_loc']]
    }
    assert len(prediction_metadata['id_str']) == len(predictions)

    galaxy_id_df = pd.DataFrame(data=prediction_metadata)

    return galaxy_id_df, predictions, template_label_cols


def prediction_hdf5_to_summary_parquet(hdf5_loc: str, save_loc: str, schema: schemas.Schema, debug=False):
    """
    Take .hdf5 file with galaxy ids and concentrations predicted by Zoobot
    Create data-release-style .parquet catalogs with vote fractions and uncertainties
    Saves two tables: 
        'friendly', with just vote fracs and nans where answers aren't relevent
        'advanced', with all vote fractions and 90% confidence intervals on those fractions
    
    Args:
        hdf5_loc (str): _description_
        save_loc (str): _description_
    """
    assert isinstance(hdf5_loc, str)

    label_cols = schema.label_cols

    # concentrations will be of (galaxy, question, model, forward_pass) after going through c_group
    # may be only one model but will still have that dimension (e.g. 1000, 39, 1, 5)
    galaxy_id_df, concentrations, _ = load_hdf5s(hdf5_loc)

    if debug:
        logging.warning('Using debug mode of 100k only')
        concentrations = concentrations[:100000]
        galaxy_id_df = galaxy_id_df[:100000]
        save_loc = save_loc.replace('.parquet', '_debug.parquet')

    # applies to all questions at once
    # hopefully also supports 3D concentrations (galaxy/question/model/pass)
    logging.info('Concentrations: {}'.format(concentrations.shape))

    # supports (galaxy, question, distribution...) shape, no TF needed
    unmasked_fractions = stats.expected_value_of_dirichlet_mixture(concentrations, schema)

    prob_of_asked = []
    for question in schema.questions:
        # human fractions imply you'd get this many votes for this question
        # vote fracs are now over all distributions, and therefore so will be expected votes
        question_prob_of_asked = stats.get_expected_votes_ml(unmasked_fractions, question, 1, schema, round_votes=False)
        prob_of_asked.append(question_prob_of_asked)

    prob_of_asked = np.stack(prob_of_asked, axis=1)

    assert len(galaxy_id_df) == len(unmasked_fractions)
    assert len(galaxy_id_df) == len(prob_of_asked)

    # for the friendly table, will mask out to only use proportion-volunteers-asked > 0.5

    # prob of asked is by question i.e. (galaxy, question) shape, so for np.where, need to duplicate out to answers
    prob_of_asked_by_answer = []
    for question_n, question in enumerate(schema.questions):
        prob_of_asked_duplicated = np.stack([prob_of_asked[:, question_n]] * len(question.answers), axis=1)  # N repetition in new axis 1
        prob_of_asked_by_answer.append(prob_of_asked_duplicated)
    prob_of_asked_by_answer = np.concatenate(prob_of_asked_by_answer, axis=1)  # concat along that ax, now (galaxy, answer) shape

    masked_fractions = np.where(prob_of_asked_by_answer >= 0.5, unmasked_fractions, np.zeros_like(unmasked_fractions) * np.nan)

    # this doesn't do super well with large memory reqs, so chunk it up
    all_lower_edges = []
    all_upper_edges = []
    chunk_size = min(len(concentrations), 10000)
    
    num_chunks = int(len(concentrations) / chunk_size)
    logging.info('{} chunks, of typical size {}'.format(num_chunks, chunk_size))

    for concentrations_chunk in np.array_split(concentrations, num_chunks):
        lower_edges, upper_edges = stats.get_confidence_intervals(concentrations_chunk, schema, interval_width=.9, gridsize=100)  
        assert len(lower_edges) == len(concentrations_chunk)
        assert len(upper_edges) == len(concentrations_chunk)
        all_lower_edges.append(lower_edges)
        all_upper_edges.append(upper_edges)
    all_lower_edges = np.concatenate(all_lower_edges)
    all_upper_edges = np.concatenate(all_upper_edges)

    # define useful cols
    fraction_cols = [col + '_fraction' for col in label_cols]
    lower_edge_cols = [col + '_90pc-lower' for col in label_cols]
    upper_edge_cols = [col + '_90pc-upper' for col in label_cols]
    proportion_asked_cols = [col + '_proportion-asked' for col in label_cols]

    # make friendly dataframe with just masked fraction and description string
    friendly_loc = save_loc.replace('.parquet', '_friendly.parquet')
    fraction_df = pd.DataFrame(data=masked_fractions, columns=fraction_cols)
    friendly_df = pd.concat([galaxy_id_df, fraction_df], axis=1)
    friendly_df.to_parquet(friendly_loc, index=False)
    logging.info('Friendly summary table saved to {}'.format(friendly_loc))

    # make advanced dataframe with unmasked fractions, error bars, proportion_asked
    advanced_loc = save_loc.replace('.parquet', '_advanced.parquet')
    fraction_df = pd.DataFrame(data=unmasked_fractions, columns=fraction_cols)
    lower_edge_df = pd.DataFrame(data=all_lower_edges, columns=lower_edge_cols)
    upper_edge_df = pd.DataFrame(data=all_upper_edges, columns=upper_edge_cols)
    proportion_df = pd.DataFrame(data=prob_of_asked_by_answer, columns=proportion_asked_cols)
    advanced_df = pd.concat([galaxy_id_df, fraction_df, lower_edge_df, upper_edge_df, proportion_df], axis=1)
    advanced_df.to_parquet(advanced_loc, index=False)
    logging.info('Advanced summary table saved to {}'.format(advanced_loc))


def single_forward_pass_hdf5s_to_df(hdf5_locs: List, drop_extra_dims=False):
    """
    Load predictions (or representations) saved as hdf5 into pd.DataFrame with id_str and label_cols columns

    Best used when not using test-time dropout i.e. when you have a single forward pass per galaxy. 
    Otherwise, with test-time dropout on, aanswer columns (keyed by each value in label_cols) will be
    `np.ndarray` of length (n forward passes) for each galaxy and answer, *not* 1D scalars.
    This is quite awkward to work with. I suggest using `load_hdf5s` directly when using test-time dropout

    Args:
        hdf5_locs (List): _description_

    Returns:
        _type_: _description_
    """
    galaxy_id_df, predictions, label_cols = load_hdf5s(hdf5_locs)

    predictions = predictions.squeeze()
    
    if len(predictions.shape) > 2:
        if drop_extra_dims:
            predictions = predictions[:, :, 0]
            logging.warning('Dropped extra dimensions')
        else:
            logging.critical(
                'Predictions are of shape {}, greater than rank 2. \
                I suggest using load_hdf5s directly to work with np.arrays, not with DataFrame - see docstring'
            )
    prediction_df = pd.DataFrame(data=predictions, columns=label_cols)
    # copy over metadata (indices will align)
    prediction_df['id_str'] = galaxy_id_df['id_str']
    prediction_df['hdf5_loc'] = galaxy_id_df['hdf5_loc']
    return prediction_df


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    # simple tests

    # load_hdf5s('/nvme1/scratch/walml/repos/understanding_galaxies/scaled_image_predictions.hdf5')
    df = single_forward_pass_hdf5s_to_df('/nvme1/scratch/walml/repos/understanding_galaxies/scaled_image_predictions.hdf5')
    print(df)



from torch.utils.tensorboard import SummaryWriter
from torch import Tensor
import torch
from pytorch_lightning.callbacks import Callback

from typing import Any, Callable, Dict, List, Optional


TRAINING_MODE: str = "training"
VALIDATION_MODE: str = "validation"
TEST_MODE: str = "test"
MODES: List[str] = [TRAINING_MODE, VALIDATION_MODE, TEST_MODE]


def prepare_log_metrics(prediction: Tensor,
                        ground_truth: Tensor,
                        criterions: List[Callable],
                        mode: str) -> Dict[str, float]:
    metrics: Dict[str, float] = {}
    for criterion in criterions:
        metrics[f'{mode}/{criterion.__name__}'] = criterion(prediction, ground_truth)
    return metrics


class MetricsWriter(Callback):
    def __init__(self,
                 writer: SummaryWriter,
                 criterions: Optional[List[Callable]] = None,
                 mode: str = TRAINING_MODE):
        """
        Args:
            writer (SummaryWriter): Tensorboard SummaryWriter object
            criterions (Optional[List[Callable]], optional): List of metric functions to log. Defaults to None.
            mode (str, optional): Should be "training" or "validation" or "test". Defaults to "training".
        """
        if mode not in MODES:
            raise ValueError("Mode must be one of 'training', 'validation', or 'test'")
        
        self.__writer = writer
        self.__criterions = criterions if criterions else []
        self.__batch_value_sum = {'loss': 0.0}
        for criterion in self.__criterions:
            self.__batch_value_sum[criterion.__name__] = 0.0
        
        self.__batches_counted = 0
        self.__training_epochs_logged = 0
        self.__mode = mode
        
    def __zero_batch_metrics(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        for k, _ in self.__batch_value_sum.items():
            self.__batch_value_sum[k] = 0.0
        self.__batches_counted = 0
        self.__training_epochs_logged += 1
        
    def __log_epoch_metrics(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        for k, v in self.__batch_value_sum.items():
            self.__writer.add_scalar(tag=f'{self.__mode}/mean_batch_{k}',
                                    scalar_value=v/self.__batches_counted,
                                    global_step=self.__training_epochs_logged)
            
    def __log_batch_metrics(
            self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", outputs: Any, batch: Any, batch_idx: int
        ) -> None:
        if type(outputs) == dict:
            loss = outputs['loss']
        elif type(outputs) == torch.Tensor and outputs.shape==():
            loss = outputs.item()
        else:
            # TODO: add warning via logging
            return
        
        self.__batch_value_sum['loss'] += loss
        self.__batches_counted += 1
        
        prediction = pl_module.predict_step(batch, batch_idx)
        for criterion in self.__criterions:
            self.__batch_value_sum[criterion.__name__] += criterion(prediction, batch[1])
        
    def on_train_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Called when the train epoch begins."""
        if self.__mode == TRAINING_MODE:
            self.__zero_batch_metrics(trainer = trainer,
                                      pl_module = pl_module)
            
    def on_validation_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Called when the train epoch begins."""
        if self.__mode == VALIDATION_MODE:
            self.__zero_batch_metrics(trainer = trainer,
                                      pl_module = pl_module)
    
    def on_test_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Called when the train epoch begins."""
        if self.__mode == TEST_MODE:
            self.__zero_batch_metrics(trainer = trainer,
                                      pl_module = pl_module)
        
    def on_train_epoch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Called when the train epoch begins."""
        if self.__mode == TRAINING_MODE:
            self.__log_epoch_metrics(trainer = trainer,
                                     pl_module = pl_module)
            
    def on_validation_epoch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Called when the train epoch begins."""
        if self.__mode == VALIDATION_MODE:
            self.__log_epoch_metrics(trainer = trainer,
                                     pl_module = pl_module)
            
    def on_test_epoch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Called when the train epoch begins."""
        if self.__mode == TEST_MODE:
            self.__log_epoch_metrics(trainer = trainer,
                                     pl_module = pl_module)
        
    def on_train_batch_end(
            self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", outputs: Any, batch: Any, batch_idx: int
        ) -> None:
        if self.__mode == TRAINING_MODE:
            self.__log_batch_metrics(trainer = trainer,
                                     pl_module = pl_module,
                                     outputs = outputs,
                                     batch = batch,
                                     batch_idx = batch_idx)
            
    def on_validation_batch_end(
            self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", outputs: Any,
            batch: Any, batch_idx: int, dataloader_idx: int
        ) -> None:
        if self.__mode == VALIDATION_MODE:
            self.__log_batch_metrics(trainer = trainer,
                                     pl_module = pl_module,
                                     outputs = outputs,
                                     batch = batch,
                                     batch_idx = batch_idx)
            
    def on_test_batch_end(
            self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", outputs: Any,
            batch: Any, batch_idx: int, dataloader_idx: int
        ) -> None:
        if self.__mode == TEST_MODE:
            self.__log_batch_metrics(trainer = trainer,
                                     pl_module = pl_module,
                                     outputs = outputs,
                                     batch = batch,
                                     batch_idx = batch_idx)



import logging
import os
import time

import tensorflow as tf

# from zoobot.tensorflow.estimators import custom_callbacks

# similar style to PyTorch Lightning
class Trainer():
    """Lighting-esque object for executing training.

    Args:
        epochs (int, optional): Max epochs to train. Defaults to 1500.
        patience (int, optional): Max epochs to wait for loss improvement before cancelling training. Defaults to 10.
        log_dir (str, optional): Directory to save training logs. Defaults to 'runs/default_run_{}'.format(time.time()).
        save_freq (str, optional): Frequency with which to save training logs. Defaults to 'epoch'.
    """
    
    def __init__(
        # this doesn't really need to be a class, 
        # but it's kinda nice to break up the training instructions from the model/dataset
            self,
            epochs=1500,  # rely on earlystopping callback
            min_epochs=0,
            patience=10,
            log_dir='runs/default_run_{}'.format(time.time()),
            save_freq='epoch'
    ): 
        self.epochs = epochs
        self.min_epochs = min_epochs
        self.patience = patience
        self.log_dir = log_dir
        self.save_freq = save_freq


    # TODO move to shared utilities
    def asdict(self):
        excluded_keys = ['__dict__', '__doc__', '__module__', '__weakref__']
        return dict([(key, value) for (key, value) in self.__dict__.items() if key not in excluded_keys])


    def fit(self, model: tf.keras.Model, train_dataset, val_dataset, test_dataset=None, extra_callbacks=[], eager=False, verbose=2):
        """
        Train and evaluate a model.

        Includes tensorboard logging (to log_dir/tensorboard).
        Includes checkpointing (named log_dir/checkpoint), with the rolling best val loss checkpoint saved.
        Includes early stopping according to train_config.patience.

        Args:
            model (tf.keras.Model): model to train. Must already be compiled with model.compile(loss, optimizer)
            train_dataset (tf.data.Dataset): yielding batched tuples of (galaxy images, labels)
            val_dataset (tf.data.Dataset): yielding batched tuples of (galaxy images, labels)
            extra_callbacks (list): any extra callbacks to use when training the model. See e.g. tf.keras.callbacks.
            eager (bool, optional): If True, train in eager mode - slow, but helpful for debugging. Defaults to False.
            verbose (int, optional): 1 for progress bar, useful for local training. 2 for one line per epoch, useful for scripts. Defaults to 2.

        Returns:
            None
        """

        if not os.path.isdir(self.log_dir):
            os.mkdir(self.log_dir)

        # will create a multi-file checkpoint like {checkpoint.index, checkpoint.data.00000-00001, ...}
        checkpoint_name = os.path.join(self.log_dir, 'checkpoint')

        tensorboard_dir = os.path.join(self.log_dir, 'tensorboard')
        tensorboard_callback = tf.keras.callbacks.TensorBoard(
                log_dir=tensorboard_dir,
                # explicitly disable various slow logging options - enable these if you like
                # https://www.tensorflow.org/api_docs/python/tf/keras/callbacks/TensorBoard
                histogram_freq=0,  # don't log all the internal histograms, possibly slow
                write_images=False,  # this actually writes the weights, terrible name
                write_graph=False,
                # profile_batch='2,10',
                profile_batch=0,   # i.e. disable profiling
                update_freq=100  # must be an integer, else logging within define_model.py will fail
            )
        checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
                filepath=checkpoint_name,
                monitor='val_loss',
                mode='min',
                save_freq=self.save_freq,
                save_best_only=True,
                save_weights_only=True)
            
        callbacks = [
            tensorboard_callback,
            checkpoint_callback,
            tf.keras.callbacks.EarlyStopping(restore_best_weights=True, patience=self.patience),
            tf.keras.callbacks.TerminateOnNaN()
            # custom_callbacks.VisualizeImages()
        ] + extra_callbacks
        # TODO https://www.tensorflow.org/api_docs/python/tf/keras/callbacks/ReduceLROnPlateau

        # https://www.tensorflow.org/tensorboard/scalars_and_keras
        # automatically logs (train/validation)/epoch_loss
        # adds into tensorboard_dir, which also has train/val subfolders
        fit_summary_writer = tf.summary.create_file_writer(os.path.join(tensorboard_dir, 'writer'))
        # pylint: disable=not-context-manager
        with fit_summary_writer.as_default(): 
            # pylint: enable=not-context-manager
            # for debugging
            if eager:
                logging.warning('Running in eager mode')
                model.run_eagerly = True
            # https://www.tensorflow.org/api_docs/python/tf/keras/Model

            # import numpy as np
            # tf.summary.image(name='example', data=np.random.rand(16, 64, 64, 1), step=0)

            # print(callbacks)
            # exit()

            model.fit(
                train_dataset,
                validation_data=val_dataset,
                epochs=self.epochs,
                callbacks=callbacks,
                verbose=verbose
                # batch_size not needed when providing tf.data.Datasets
            )


        logging.info('All epochs completed - finishing gracefully')
        # note that the BEST model is saved as the latest checkpoint, but self.model is the LAST model after training completes
        # to set self.model to the best model, load the latest checkpoint 
        logging.info('Loading and returning (best) model')
        model.load_weights(checkpoint_name)  # inplace

        if test_dataset is not None:
            logging.info('Evaluating on test dataset - be careful not to overfit your choices')
            metrics = model.evaluate(
                test_dataset,
                callbacks=[tensorboard_callback],
                verbose=verbose
            )
            logging.info(metrics)
            logging.info(model.metrics_names)
            import wandb
            # wandb.log(dict(zip(model.metrics_names, metrics)))
            wandb.log({'test/epoch_loss': metrics})
        else:
            logging.info('Skipping test evaluation')

        return model



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zoobot",
    version="1.0.1",
    author="Mike Walmsley",
    author_email="walmsleymk1@gmail.com",
    description="Galaxy morphology classifiers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mwalmsley/zoobot",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: GPU :: NVIDIA CUDA"
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.8",  # recommend 3.9 for new users. TF needs >=3.7.2, torchvision>=3.8
    extras_require={
        'pytorch_cpu': [
            # A100 GPU currently only seems to support cuda 11.3 on manchester cluster, let's stick with this version for now
            # very latest version wants cuda 11.6
            'torch == 1.12.1+cpu',
            'torchvision == 0.13.1+cpu',
            'torchaudio == 0.12.1',
            'pytorch-lightning >= 2.0.0',
            'simplejpeg',
            'albumentations',
            'pyro-ppl == 1.8.0',
            'torchmetrics == 0.11.0',
            'timm'
        ],
        'pytorch_m1': [
            # as above but without the +cpu (and the extra-index-url in readme has no effect)
            # all matching pytorch versions for an m1 system will be cpu
            'torch == 1.12.1',
            'torchvision == 0.13.1',
            'torchaudio == 0.12.1',
            'pytorch-lightning >= 2.0.0',
            'simplejpeg',
            'albumentations',
            'pyro-ppl == 1.8.0',
            'torchmetrics == 0.11.0',
            'timm'
        ],
        # as above but without pytorch itself
        # for GPU, you will also need e.g. cudatoolkit=11.3, 11.6
        # https://pytorch.org/get-started/previous-versions/#v1121
        'pytorch_cu113': [
            'torch == 1.12.1+cu113',
            'torchvision == 0.13.1+cu113',
            'torchaudio == 0.12.1',
            'pytorch-lightning >= 2.0.0',
            'simplejpeg',
            'albumentations',
            'pyro-ppl == 1.8.0',
            'torchmetrics == 0.11.0',
            'timm'
        ],
        'pytorch_colab': [
            'pytorch-lightning >= 2.0.0',
            'simplejpeg',
            'albumentations',
            'pyro-ppl>=1.8.0',
            'torchmetrics==0.11.0',
            'timm'
        ],
        'tensorflow': [
            'tensorflow == 2.10.0',  # 2.11.0 turns on XLA somewhere which then fails on multi-GPU...TODO
            'keras_applications',
            'tensorflow_probability == 0.18.0',  # 0.19 requires tf 2.11
            'protobuf <= 3.19'  # tensorflow incompatible above this (usually resolved by pip automatically)
        ],
        # for GPU, you will also need cudatoolkit=11.2 and cudnn=8.1.0 (note - 11.3 NOT supported by TF)
        # https://www.tensorflow.org/install/pip#step-by-step_instructions
        'utilities': [
            'seaborn',  # for nice plots
            'boto3',    # for AWs s3 access
            'python-dateutil == 2.8.1',  # for boto3  
        ],
        'docs': [
            'Sphinx',
            'sphinxcontrib-napoleon',
            'furo',
            'docutils<0.18'
        ]
    },
    install_requires=[
        'h5py',
        'tqdm',
        'pillow',
        'numpy',
        'pandas',
        'scipy',
        'astropy',  # for reading fits
        'scikit-image >= 0.19.2',
        'scikit-learn >= 1.0.2',
        'matplotlib',
        'pyarrow',  # to read parquet, which is very handy for big datasets
        # for saving metrics to weights&biases (cloud service, free within limits)
        'wandb',
        'setuptools==59.5.0',  # wandb logger incompatibility
        'galaxy-datasets==0.0.11'  # for dataset loading in both TF and Torch (renamed from pytorch-galaxy-datasets)
    ]
)
